""" feats_squim.py
predict SQUIM ( SPEECH QUALITY AND INTELLIGIBILITY
MEASURES) features


    Wideband Perceptual Estimation of Speech Quality (PESQ) [2]
    Short-Time Objective Intelligibility (STOI) [3]
    Scale-Invariant Signal-to-Distortion Ratio (SI-SDR) [4]


adapted from
from https://pytorch.org/audio/main/tutorials/squim_tutorial.html#sphx-glr-tutorials-squim-tutorial-py
paper: https://arxiv.org/pdf/2304.01448.pdf

needs 
pip uninstall -y torch torchvision torchaudio
pip install --pre torch torchvision torchaudio --extra-index-url https://download.pytorch.org/whl/nightly/cpu

"""
from nkululeko.util import Util
from nkululeko.featureset import Featureset
import os
import pandas as pd
import os
import nkululeko.glob_conf as glob_conf
import audiofile
import torch
import torchaudio
from torchaudio.pipelines import SQUIM_OBJECTIVE

class SQUIMSet(Featureset):
    """Class to predict SQUIM features
    
    """

    def __init__(self, name, data_df):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df)
        self.device = self.util.config_val('MODEL', 'device', 'cpu')
        self.model_initialized = False


    def init_model(self):
        # load model
        self.util.debug('loading model...')
        self.objective_model = SQUIM_OBJECTIVE.get_model()
        self.model_initialized = True


    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path('store')
        store_format = self.util.config_val('FEATS', 'store_format', 'pkl')
        storage = f'{store}{self.name}.{store_format}'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug('predicting SQUIM, this might take a while...')
            emb_series = pd.Series(index = self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(self.data_df.index.to_list()):
                signal, sampling_rate = audiofile.read(file, offset=start.total_seconds(), duration=(end-start).total_seconds(), always_2d=True)
                emb = self.get_embeddings(signal, sampling_rate)
                emb_series[idx] = emb
                if idx%10==0:
                    self.util.debug(f'SQUIM: {idx} of {length} done')
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index) 
            self.df.columns = ['pesq', 'sdr', 'stoi']
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'false'
            except KeyError:
                pass
        else:
            self.util.debug('reusing predicted SQUIM values')
            self.df = self.util.get_store(storage, store_format)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                print(nanrows)
                self.util.error(f'got nan: {self.df.shape} {self.df.isnull().sum().sum()}')


    def get_embeddings(self, signal, sampling_rate):
        tmp_audio_name = 'squim_audio_tmp.wav'
        audiofile.write(tmp_audio_name, signal, sampling_rate)
        WAVEFORM_SPEECH, SAMPLE_RATE_SPEECH = torchaudio.load(tmp_audio_name)
        with torch.no_grad():
            stoi_hyp, pesq_hyp, si_sdr_hyp = self.objective_model(WAVEFORM_SPEECH)
        pesq = float(pesq_hyp[0].numpy())
        stoi = float(stoi_hyp[0].numpy())
        sdr = float(si_sdr_hyp[0].numpy())
        return pesq, sdr, stoi
 
    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr)
        return feats


# segment.py
# segment a database

from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from  nkululeko.constants import VERSION
import argparse
import os
import torch
import audformat
from audformat.utils import to_filewise_index
from audformat import segmented_index
import pandas as pd 

# initialize the VAD model
SAMPLING_RATE = 16000
torch.set_num_threads(1)
vad_model, vad_utils = torch.hub.load(repo_or_dir='snakers4/silero-vad',
                    model='silero_vad',
                    force_reload=False,
                    onnx=False)
(get_speech_timestamps,
save_audio,
read_audio,
VADIterator,
collect_chunks) = vad_utils

def main(src_dir):
    parser = argparse.ArgumentParser(description='Call the nkululeko framework.')
    parser.add_argument('--config', default='exp.ini', help='The base configuration')
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config    
    else:
        config_file = f'{src_dir}/exp.ini'

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f'ERROR: no such file: {config_file}')
        exit()
        
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util('segment')
    util.debug(f'running {expr.name}, nkululeko version {VERSION}')

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f'train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}')

    # segment
    segment_target = util.config_val('DATA', 'segment_target', '_seg')
    # this if a specific dataset is to be segmented
    # segment_db = util.config_val('DATA', 'segment', False)
    # if segment_db:
    #     for dataset in expr.datasets.keys:
    #         if segment_db == dataset:
    #             df = expr.datasets[dataset].df
    #             util.debug(f'segmenting {dataset}') 
    #             df_seg = segment_dataframe(df) 
    #             name = f'{dataset}{segment_target}'
    #             df_seg.to_csv(f'{expr.data_dir}/{name}.csv')

    sample_selection = util.config_val('DATA', 'segment', 'all')
    if sample_selection=='all':
        df = pd.concat([expr.df_train, expr.df_test])
    elif sample_selection=='train':
        df = expr.df_train
    elif sample_selection=='test':
        df = expr.df_test
    else:
        util.error(f'unknown segmentation selection specifier {sample_selection}, should be [all | train | test]')

    if "duration" not in df.columns:
        df = df.drop(columns=['duration'], inplace=True)
    util.debug(f'segmenting train and test set: {df.shape[0]} samples') 
    df_seg = segment_dataframe(df) 
    def calc_dur(x):
        starts = x[1]
        ends = x[2]
        return (ends - starts).total_seconds()
    df_seg['duration'] = df_seg.index.to_series().map(lambda x:calc_dur(x)) 
    dataname = '_'.join(expr.datasets.keys())
    name = f'{dataname}{segment_target}'
    df_seg.to_csv(f'{expr.data_dir}/{name}.csv')
    print('')
    util.debug(f'saved {name}.csv to {expr.data_dir}, {df_seg.shape[0]} samples')
    print('DONE')




def get_segmentation(file):
#    print(f'segmenting {file[0]}')
    print('.', end='')
    wav = read_audio(file[0], sampling_rate=SAMPLING_RATE)
    speech_timestamps = get_speech_timestamps(wav, vad_model, sampling_rate=SAMPLING_RATE)
    files, starts, ends = [], [], []
    for entry in speech_timestamps:
        start = float(entry['start']/1000.0)
        end = float(entry['end']/1000.0)
        files.append(file[0])
        starts.append(start)
        ends.append(end)
    seg_index = segmented_index(files, starts, ends)
    return seg_index

def segment_dataframe(df):
    dfs = []
    for file, values in df.iterrows():
        index = get_segmentation(file)
        dfs.append(
            pd.DataFrame(
                values.to_dict(),
                index,
            )
        )
    return audformat.utils.concat(dfs)


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd) # use this if you want to state the config file path on command line


""""
A predictor for biological sex.
Currently based on audEERING's agender model.
"""
from nkululeko.util import Util
from nkululeko.feature_extractor import FeatureExtractor
import ast
import nkululeko.glob_conf as glob_conf
class GenderPredictor:
    """
    GenderPredictor
    predicting gender with the audEERING agender model
    
    """
    def __init__(self, df):
        self.df = df
        self.util = Util('genderPredictor')


    def predict(self, split_selection):
        self.util.debug(f'predicting gender for {split_selection} samples')
        feats_name = "_".join(ast.literal_eval(glob_conf.config['DATA']['databases']))
        self.feature_extractor = FeatureExtractor(self.df, ['agender_agender'], feats_name, split_selection)
        agender_df = self.feature_extractor.extract()
        pred_gender = agender_df.drop('age', axis=1).idxmax(axis=1)
        return_df = self.df.copy()
        return_df['gender_pred'] = pred_gender
        return return_df




""""
A predictor for SNR - signal-to-noise ratio.
"""
from nkululeko.util import Util
import ast
import nkululeko.glob_conf as glob_conf
from nkululeko.feature_extractor import FeatureExtractor
import numpy as np


class SNRPredictor:
    """
    SNRPredictor
    predicting snr
    
    """
    def __init__(self, df):
        self.df = df
        self.util = Util('snrPredictor')


    def predict(self, split_selection):
        self.util.debug(f'estimating SNR for {split_selection} samples')
        return_df = self.df.copy()
        feats_name = "_".join(ast.literal_eval(glob_conf.config['DATA']['databases']))
        self.feature_extractor = FeatureExtractor(self.df, ['snr'], feats_name, split_selection)
        result_df = self.feature_extractor.extract()
        # replace missing values by 0
        result_df = result_df.fillna(0)
        result_df = result_df.replace(np.nan, 0)
        result_df.replace([np.inf, -np.inf], 0, inplace=True)
        pred_snr = result_df.snr * 100
        return_df['snr_pred'] = pred_snr.astype('int')/100
        return return_df




""" feats_snr.py
    Estimate snr (signal to noise ratio as acoustic features)
"""
from nkululeko.util import Util
from nkululeko.featureset import Featureset
from nkululeko.estimate_snr import SNREstimator
import os
import pandas as pd
import os
import nkululeko.glob_conf as glob_conf
import audiofile

class SNRSet(Featureset):
    """Class to estimate snr"""

    def __init__(self, name, data_df):
        """Constructor. """
        super().__init__(name, data_df)

    def extract(self):
        """Estimate the features or load them from disk if present."""
        store = self.util.get_path('store')
        store_format = self.util.config_val('FEATS', 'store_format', 'pkl')
        storage = f'{store}{self.name}.{store_format}'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            self.util.debug('estimating SNR, this might take a while...')
            snr_series = pd.Series(index = self.data_df.index, dtype=object)
            for idx, (file, start, end) in enumerate(self.data_df.index.to_list()):
                signal, sampling_rate = audiofile.read(file, offset=start.total_seconds(), duration=(end-start).total_seconds(), always_2d=True)
                snr = self.get_snr(signal[0], sampling_rate)
                snr_series[idx] = snr
                if idx%10==0:
                    print('.', end='')
            print('')
            self.df = pd.DataFrame(snr_series.values.tolist(), index=self.data_df.index) 
            self.df.columns = ['snr']
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config['DATA']['needs_feature_extraction'] = 'false'
            except KeyError:
                pass
        else:
            self.util.debug('reusing estimated SNR values')
            self.df = self.util.get_store(storage, store_format)

    def get_snr(self, signal, sampling_rate):
        r"""Estimate SNR from raw audio signal.
        Args:
            signal: audio signal
            sampling_rate: sample rate
        Returns
            snr: estimated signal to noise ratio
        """
        snr_estimator = SNREstimator(signal, sampling_rate)
        estimated_snr, log_energies, energy_threshold_low, energy_threshold_high = snr_estimator.estimate_snr()
        return estimated_snr
    
    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_snr(signal, sr)
        return feats


# feats_oxbow.py

from nkululeko.util import Util
from nkululeko.featureset import Featureset
import os
import pandas as pd
import opensmile

class Openxbow(Featureset):
    """Class to extract openXBOW processed opensmile features (https://github.com/openXBOW)"""

    def __init__(self, name, data_df, is_train = False):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df)
        self.is_train = is_train

    def extract(self):
        """Extract the features or load them from disk if present."""
        self.featset = self.util.config_val('FEATS', 'set', 'eGeMAPSv02')
        self.feature_set = eval(f'opensmile.FeatureSet.{self.featset}')
        store = self.util.get_path('store')
        storage = f'{store}{self.name}_{self.featset}.pkl'
        extract = self.util.config_val('FEATS', 'needs_feature_extraction', False)
        no_reuse = eval(self.util.config_val('FEATS', 'no_reuse', 'False'))
        if extract or no_reuse or not os.path.isfile(storage):
            # extract smile features first
            self.util.debug('extracting openSmile features, this might take a while...')
            smile = opensmile.Smile(
                feature_set= self.feature_set,
                feature_level=opensmile.FeatureLevel.LowLevelDescriptors,
                num_workers=5,)
            if isinstance(self.data_df.index, pd.MultiIndex):
                is_multi_index = True
                smile_df = smile.process_index(self.data_df.index)
            else:
                smile_df = smile.process_files(self.data_df.index)
            smile_df.index = smile_df.index.droplevel(1)
            smile_df.index = smile_df.index.droplevel(1)
            # compute xbow features
            # set some file names on disk
            lld_name, xbow_name, codebook_name = 'llds.csv', 'xbow.csv', 'xbow_codebook'
            # save the smile features
            smile_df.to_csv(lld_name, sep=';', header=False)
            # get the path of the xbow java jar file
            xbow_path = self.util.config_val('FEATS', 'xbow.model', '../openXBOW/')
            # get the size of the codebook
            size = self.util.config_val('FEATS', 'size', 500)
            # get the number of assignements
            assignments = self.util.config_val('FEATS', 'assignments', 10)
            # differentiate between train and test
            if self.is_train:
                # store the codebook
                os.system(f'java -jar {xbow_path}openXBOW.jar -i {lld_name} -standardizeInput -log \
                    -o {xbow_name} -size {size} -a {assignments} -B {codebook_name}')
            else:
                # use the codebook
                os.system(f'java -jar {xbow_path}openXBOW.jar -i {lld_name} \
                    -o {xbow_name} -b {codebook_name}')
            # read in the result from disk
            xbow_df = pd.read_csv(xbow_name, sep=';', header=None)
            # set the index
            xbow_df = xbow_df.set_index(self.data_df.index)
            # check if smile features should be added
            with_os = self.util.config_val('FEATS', 'with_os', False)
            if with_os:
                # extract smile functionals
                self.util.debug('extracting openSmile functionals, this might take a while...')
                smile = opensmile.Smile(
                    feature_set= opensmile.FeatureSet.eGeMAPSv02, # always use eGemaps for this
                    feature_level=opensmile.FeatureLevel.Functionals,
                    num_workers=5,)
                if isinstance(self.data_df.index, pd.MultiIndex):
                    is_multi_index = True
                    smile_df = smile.process_index(self.data_df.index)
                else:
                    smile_df = smile.process_files(self.data_df.index)
                # drop the multi index
                smile_df.index = smile_df.index.droplevel(1)
                smile_df.index = smile_df.index.droplevel(1)
                xbow_df = xbow_df.join(smile_df)
            # in any case, store to disk for later use
            xbow_df.to_pickle(storage) 
            # and assign to be the "official" feature set
            self.df = xbow_df           
        else:
            self.util.debug('reusing extracted OS features.')
            self.df = pd.read_pickle(storage)
